// MacGuffin Sequencer_V1-06 18 FEB 2022
// MacGuffin sequencer for Portsdown
// Arduino Nano A000005 or clone

#include <TimerOne.h>   // timer tick reference
#include <Wire.h>       // I2C
#include <LiquidCrystal_I2C.h>  // from library manager

const byte IOlcd  = 0x3F;  // LCD  address

//       1 = enables IO expander
//       0 = turns off error check for uninstalled IO expanders
#define Exp_21H  1     // IO expander 1
#define Exp_22H  1     // IO expander 2
#define Exp_23H  0     // IO expander 3
#define Exp_24H  0     // IO expander 4
#define Exp_25H  0     // IO expander 5
#define Exp_26H  0     // IO expander 6
#define Exp_27H  0     // IO expander 7

#define PPoff 1   // 1 = enable PA power off delay

#define RCVR  0   // enables receiver band input
#define RVinv 0   // 0 = band input from receiver is used or
                  // hex switch is used without inverter IC 
                  // 1 =  hex switch is used with inverter IC
 
      // Timer increment is 25 mS, i.e. 40 = 1 second 
const int R2Tdly = 40;   // delay from receive to transmit
const int T2Rdly = 40;   // delay from transmit to receive
const int PPdly  = 300; // off delay of PA DC power after TX ends
                         // 13.6 minutes max ( 32767 )
const int anaCal = 1300; // analog calibration (1250)
//if analog input of 1024 = 5V then V*1000 = (analog*1250)/256

// The following 4 items are the Portsdown table, it holds
// LCD band text ( must be 7 characters)
// Transverter used
// and links RX and TX Path #s to Portsdown scenario #

// LCD Band text for Portsdown scenarios
const char *pdLCD[16] = { 
  " 71 MHz" , "146 MHz" , "  70 cm" , "  23 cm" , //band 0-3
  "   9 cm" , "   6 cm" , "   3 cm" , " 24 GHz" , //band 4-7
  " 50 MHz" , " 47 GHz" , " 76 GHz" , "  13 cm" , //band 8-11
  " 50 MHz" ,  "  9 cm" , "  Spare" , "  13 cm"   //band 12-15
};

// 0 = no transverter, 1 to 9 show on LCD when scenario is selected.
const byte pdTV[16] = {       // transverter    0 = not used
// 0***1***2***3***4***5***6***7***8***9**10**11**12**13**14**15**band
   0,  0,  0,  0,  1,  1,  1,  1,  0,  1,  1,  0,  1,  0,  1,  1
};

// RX path # linked to Portsdown scenario
const byte pdRX[16] = {       // RX path  16 = not used
// 0***1***2***3***4***5***6***7***8***9**10**11**12**13**14**15**band
   0,  1,  2,  3, 16, 16, 16, 16,  5, 10, 16,  4, 12,  6,  7, 16
};

// TX path # linked to Portsdown scenario
const byte pdTX[16] = {       // TX path  16 = not used
// 0***1***2***3***4***5***6***7***8***9**10**11**12**13**14**15**band
   0,  1,  2,  3, 16, 16, 16, 16,  5, 10, 16,  4, 12,  6,  7, 16
};

// LCD Text for RX and TX
// RX path LCD text table
const char *rxLCD[16] = {     // text for LCD
  " 71 MHz" , "146 MHz" , "  70 cm" , "  23 cm" , //band 0-3
  "  13 cm" , " 50 MHz" , "   9 cm" , "  Spare" , //band 4-7
  "   sp 8" , "   sp 9" , "   sp10" , "   sp11" , //band 8-11
  "   sp12" , "   sp13" , "   sp14" , "   sp15"   //band 12-15
};

// TX path LCD text table
const char *txLCD[16] = {     // text for LCD
  " 71 MHz" , "146 MHz" , "  70 cm" , "  23 cm" , //band 0-3
  "  13 cm" , " 50 MHz" , "   9 cm" , "  Spare" , //band 4-7
  "   sp 8" , "   sp 9" , "   sp10" , "   sp11" , //band 8-11
  "   sp12" , "   sp13" , "   sp14" , "   sp15"   //band 12-15
};

// This is the optional receiver LCD text table
const char *rvLCD[16] = {     // text for LCD
  " 71 MHz" , "146 MHz" , "  70 cm" , "  23 cm" , //band 0-3
  "  13 cm" , " 50 MHz" , "   9 cm" , "  Spare" , //band 4-7
  "   sp 8" , "   sp 9" , "   sp10" , "   sp11" , //band 8-11
  "   sp12" , "   sp13" , "   sp14" , "   sp15"   //band 12-15
};

/* ***** OUTPUT PORT DEFINITION TABLES *****
     PARAMETERS
   PATH SELECTION, TIMING
       RX-TX                        BIT-
        RX-TIME TX-TIME             MODE   OUTPUT PATTERN
    0 = RX      OFF  RX,INVERT T/R    0= 1 BIT FROM 0 TO 7
    1 = RX      RX   RX               1= 1 BIT FROM 8 TO 15
    2 = OFF     TX   TX,T/R           2= HEX VALUE OF BAND
    3 = TX      TX   TX,T/R           3= HEX VALUE OF BAND + 1
    4 = OFF     PA DC POWER           4= USER BIT PATTERN 1, LOWER 8 BITS
    5 = OFF     PA PTT                5= USER BIT PATTERN 1, UPPER 8 BITS
    6 = RX      TX                    6= USER BIT PATTERN 2, LOWER 8 BITS
                                      7= USER BIT PATTERN 2, UPPER 8 BITS
   16 = UNUSED  UNUSED               16= UNUSED

    INVERT OUTPUT PORT, 0 = OFF, 1 = ON
*/  
      //IO Definition table
const byte IOdef[16] [3] = {
  // PATH    BIT-
  // RX-TX   MODE  INVERT
  {    0,     4,      0 }, // 20H PA  IO Exp 0
  {    2,     5,      0 }, // 20H PB
  {    5,     0,      1 }, // 21H PA  IO Exp 1
  {    4,     0,      1 }, // 21H PB
  {    6,     6,      0 }, // 22H PA  IO Exp 2
  {    0,     0,      1 }, // 22H PB
  {   16,     0,      0 }, // 23H PA  IO Exp 3
  {   16,     0,      0 }, // 23H PB
  {   16,     0,      0 }, // 24H PA  IO Exp 4
  {   16,     0,      0 }, // 24H PB
  {   16,     0,      0 }, // 25H PA  IO Exp 5
  {   16,     0,      0 }, // 25H PB
  {   16,     0,      0 }, // 26H PA  IO Exp 6
  {   16,     0,      0 }, // 26H PB
  {   16,     0,      0 }, // 27H PA  IO Exp 7
  {   16,     0,      0 }  // 27H PB
};

// The table above is configured for this RF relay configuration
// IO   Port    DEFAULT USE
// Exp0 20H PA  TR select   Logic 0 to 7
// Exp0 20H PB  TX select   Logic Hex + 0
// Exp1 21H PA  PA PTT      Relay 0 to 7
// Exp1 21H PB  PA POWER    Relay 0 to 7
// Exp2 22H PA  RX select   Relay 0 to 7
// Exp2 22H PB  spare
// Exp3 to Exp8 not used

    // User defined IO bit pattern table 1
const unsigned int UDBP1[16]{
// path*0  path*1  path*2  path*3  path*4  path*5  path*6  path*7
// path*8  path*9  path10  path11  path12  path13  path14  path15

   0x9111, 0xA222, 0xB333, 0xC444, 0xD555, 0xE666, 0x000, 0x000, 
   0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 };

    // User defined IO bit pattern table 2
const unsigned int UDBP2[16]{
// path*0  path*1  path*2  path*3  path*4  path*5  path*6  path*7
// path*8  path*9  path10  path11  path12  path13  path14  path15

   0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0000, 0x0000, 
   0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 };

//        User modification area ends here
//********************************************************

// Arduino IO assignment
//const byte xxxxx = 0;     // (RX)
//const byte xxxxx = 1;     // (TX)
const byte CB0   = 2;     // in D1 Cross band
const byte CB1   = 3;     // in D2
const byte CB2   = 4;     // in D4
const byte CB3   = 5;     // in D8
const byte CBS   = 6;     // in Enable Cross band
//const byte xxxxx = 7;     //
const byte TB0   = 8;     // in D1 Portsdown
const byte TB1   = 9;     // in D2
const byte TB2   = 10;    // in D4
const byte TB3   = 11;    // in D8
const byte TTX   = 12;    // TX-ON from Portsdown
//const byte xxxxx = 13;    // (LED)
//
const byte RV0   = A0;    // in D1 receiver band
const byte RV1   = A1;    // in D2
const byte RV2   = A2;    // in D4
const byte RV3   = A3;    // in D8
//const byte xxxxx = A4     // (SDA)
//const byte xxxxx = A5     // (SLC)
const byte R3V3  = A6;    // analog receiver 3V3 check
const byte T3V3  = A7;    // analog Portsdown 3V3 check

const byte GPIOA  = 0x12;  //PORTA Reg Addr
const byte GPIOB  = 0x13;  //PORTB Reg Addr
const byte IODIRA = 0x00;  //PORTA Direction
const byte IODIRB = 0x01;  //PORTB Direction
const byte IOout  =    0;  //All out

// I2C addresses           // DEFAULT USE
const byte IOexp0 = 0x20;  // IO for RX select, spare
const byte IOexp1 = 0x21;  // IO for PA PTT, PA POWER
const byte IOexp2 = 0x22;  // IO for TR select, TX select
const byte IOexp3 = 0x23;  // IO for <future>
const byte IOexp4 = 0x24;  //
const byte IOexp5 = 0x25;  //
const byte IOexp6 = 0x26;  //
const byte IOexp7 = 0x27;  //

byte IO_P[16];            //IO output bits
byte IO_Plst[16];         //IO output bits last sent
bool IO_chg;              //IO bits have changed

byte MCPerr = 0;        // I2C error return
byte MCPlog = 0;        // log error by bit position
bool errLCD = false;    // LCD is displaying errors

bool tick25 = false;   // true every 25 mS, transfers to tick25A
bool tick25A = false;  // true one scan every 25 mS
byte  overDiv = 0;     // over timer
byte  overSec = 0;
int   overMin = 0;
byte  overSecLst = 59;
int   overMinLst = 99;

int R2Ttmr;               //receive to transmit timer
bool R2T;

int T2Rtmr;   //transmit to receive timer
bool T2R;
bool TRrly;   //T/R relay on, not receive time

int PPtmr;    // PA Power off delay timer
bool PPOhold = false;

// output timing signals
bool PTT;         // Enable driver and/or PA
bool TRout;       // T/R relay signal, inverse can be RX relay signal
bool PApwr;       // Turns on DC power to PA, same as TRout
                  // with delay off option

bool TXinp  =  0;     // input - Portsdown TX ON
bool TXchg  =  0;     // PDval changed while TX on
bool TXchgLst = 1;

//Signals used for LCD display

bool LCDchg;
bool CBsw   =  0;     // input - cross band enabled
byte CBval  =  0;     // input - cross band hex switch
byte PDval  =  0;     // input - Portsdown band bits
byte RVval  =  0;     // input - receiver band bits
byte RXinx  =  0;     // index to select RX or crossband LCD
byte TXinx  =  0;     // index to select TX
byte TVinx  =  0;     // index to select Transverter
bool CBinx;           // index to Cross band on

bool CBswLst   =  0;     // last - cross band enabled
byte CBvalLst  = 99;     // last - cross band hex switch
byte PDvalLst  = 99;     // last - Portsdown band bits
byte PDvalIn   = 99;
byte RVvalLst  = 99;     // last - receiver band bits
byte RXinxLst  = 99;     // last select RX or crossband LCD
byte TXinxLst  = 99;
byte TVinxLst  = 99;     // last selected transverter
bool CBinxLst  =  0;     // last Cross band on
bool TXinpLst  =  0;     // last TXinp
bool TXinpR    =  0;     // TXinp raw
long  L3V3;
unsigned int PD3V3 = 0;     // Portsdown 3V3, volts *1000
unsigned int RV3V3 = 0;     // receiver 3V3, volts *1000
int PDVtmr = 0;             // Portsdown voltage OK timer

LiquidCrystal_I2C lcd(IOlcd,20,4);

void setup() {
  Timer1.initialize(25000); // timer = 25ms
  Timer1.attachInterrupt( timerIsr ); // attach the service routine

  pinMode(CB0, INPUT_PULLUP);
  pinMode(CB1, INPUT_PULLUP);
  pinMode(CB2, INPUT_PULLUP);
  pinMode(CB3, INPUT_PULLUP);
  pinMode(CBS, INPUT_PULLUP);

  pinMode(TB0, INPUT_PULLUP);
  pinMode(TB1, INPUT_PULLUP);
  pinMode(TB2, INPUT_PULLUP);
  pinMode(TB3, INPUT_PULLUP);
  pinMode(TTX, INPUT_PULLUP);

  pinMode(RV0, INPUT_PULLUP);
  pinMode(RV1, INPUT_PULLUP);
  pinMode(RV2, INPUT_PULLUP);
  pinMode(RV3, INPUT_PULLUP);

  Wire.begin();         // initialize I2C link for IO expanders
  Wire.setClock(100000); //????

  lcd.init();
  lcd.backlight();

  IO_MCP_INIT();

  CycRst();       // reset T/R timing cycle
  LCDrst();       // reset LCD
  LCDbkgd();      // write LCD background
  
  get3V3();       //read analog voltage

}

void loop() {
  noInterrupts();    // prevent interrupt while transfering
  tick25A = tick25;  //tick set by interrupt, reset next scan
  tick25 = LOW;      //Transfer and clear at begining of loop
  interrupts();

  if(TRrly&&tick25A){ //over  timer
    if(++overDiv>39){
      overDiv=0;
      if(++overSec>59){
        overSec=0;
        if(++overMin>959){
          overMin=0;
        }
      }
    }
  }

  if (PD3V3 < 3000){          // Portsdown voltage OK
    PDVtmr = 0;               // reset PD 3V3 OK timer
  }
  else{
    if( (PDVtmr < 4) && tick25A){
      PDVtmr++;               // count to 4 and hold
    }
  }
  if (!MCPlog && (PDVtmr > 3)) { //run T/R sequence if no faults
    CBval = 0;  // Cross band switch
    bitWrite(CBval, 0, !digitalRead(CB0));
    bitWrite(CBval, 1, !digitalRead(CB1));
    bitWrite(CBval, 2, !digitalRead(CB2));
    bitWrite(CBval, 3, !digitalRead(CB3));
    CBsw = !digitalRead(CBS);

    if (RV3V3 >= 3000) {
      RVval = 0;   // receiver band bits
#if RVinv
      bitWrite(RVval, 0, digitalRead(RV0));
      bitWrite(RVval, 1, digitalRead(RV1));
      bitWrite(RVval, 2, digitalRead(RV2));
      bitWrite(RVval, 3, digitalRead(RV3));
#else
      bitWrite(RVval, 0, !digitalRead(RV0));
      bitWrite(RVval, 1, !digitalRead(RV1));
      bitWrite(RVval, 2, !digitalRead(RV2));
      bitWrite(RVval, 3, !digitalRead(RV3));
#endif
    }
    else {
    RVval = 99; // 99 = don't display
    }

    PDvalIn = 0;   // Portsdown band bits
    bitWrite(PDvalIn, 0, !digitalRead(TB0));
    bitWrite(PDvalIn, 1, !digitalRead(TB1));
    bitWrite(PDvalIn, 2, !digitalRead(TB2));
    bitWrite(PDvalIn, 3, !digitalRead(TB3));
    TXinpR = !digitalRead(TTX); //scan in TXinp from Pi, low = on
    if(tick25A){        // change TXinp only on tick so first
    TXinp = TXinpR ;    // timer count is a full tick
    }
    
    if(pdTX[PDval]>15) TXinp=false;  // block TX if no TX path
    
                  //scenario change while TX is on
    if(!TXinpR) TXchg = false;     //hold TXchg fault until TX off
    if( (PDval!=PDvalIn) &&TRout){
      TXchg=true;
    }
    if(TXchg){
      TXinp=false;   // hold scenario while ending TX
    }
    else{
      if(PDval!=PDvalIn) PPOhold=false; //cancel PA delay off if path changes
      PDval=PDvalIn;
      if(pdRX[PDval]>15) TXinp=false;  // unused scenario
    }
      
    TVinx = pdTV[PDval]; // setup indexs for LCD and output decode
    TXinx = pdTX[PDval];
    if (CBsw) {
      RXinx = pdRX[CBval];
    }
    else {
      RXinx = pdRX[PDval];
    }

            // T/R timing sequence
    if (!TXinp) {
      T2Rtmr = T2Rdly+1;    // preset T2R delay timer
      T2R = false;
      overDiv = 0;
      overSec = 0;
      overMin = 0;
      if (!R2T && tick25A) {   // not timed out
        --R2Ttmr;
        if (R2Ttmr < 1) {    // timed out
          R2T = true;
          TRrly = false;
        }
      }
    }

    if (TXinp) {
      R2Ttmr = R2Tdly+1;    // preset R2T delay timer
      R2T = false;
      if (!T2R && tick25A) {   // not timed out
        --T2Rtmr;
        if (T2Rtmr < 1) {    // timed out
          T2R = true;
          TRrly = true;
        }
      }
    }

#if PPoff
    if (TXinp || TRrly) {
      PPtmr = PPdly;
      PPOhold = true;
    }
    else {
#endif
      if (PPOhold && tick25A) {
        --PPtmr;
        if (PPtmr < 1) {
          PPOhold = false;
        }
      }
#if PPoff
    }
#endif
          // timing references
    TRout = TXinp || TRrly;
    PTT =   T2R;
    PApwr = TRout || PPOhold;

    IO_chg=false;
    for(byte j=0;j<16;j++){
      IO_P[j] = BitPat(j);
  if(IOdef[j] [2]) IO_P[j]=IO_P[j] ^ 0xFF; //invert
      if(IO_Plst[j]!=IO_P[j]){
        IO_Plst[j]=IO_P[j];
        IO_chg=true;
      }
    }
    
    if(IO_chg){
    IO_MCP_WR();  // send data to IO expander ports
    }

    //   *** LCD ***
    LCDchg = false;
    // PD
    if (PDvalLst != PDval) {
      PDvalLst = PDval;
      LCDchg = true;
      lcd.setCursor(6, 0);
      lcd.print(pdLCD[PDval]);   //PD band
      lcd.setCursor(3, 0);
      if ((PDval) < 10) lcd.print(" ");
      lcd.print(PDval);   //PD band
    }
    // TV
    if (TVinxLst != TVinx) {
      TVinxLst = TVinx;
      LCDchg = true;
      lcd.setCursor(0, 1);
      if (TVinx > 0) {
        lcd.print("T");
        lcd.print(TVinx);
      }
      else {
        lcd.print("TX");
      }
    }
      // TX
    if (TXinxLst != TXinx) {
      TXinxLst = TXinx;
      LCDchg = true;
      lcd.setCursor(6, 1);
      if (TXinx > 15) {
        lcd.print(" unused");   //PD band
      }
      else {
        lcd.print( txLCD[TXinx ] );   //TX band
      }
      lcd.setCursor(3, 1);
      if (TXinx < 16) {
        if ((TXinx) < 10) lcd.print(" ");
        lcd.print(TXinx);   //TX band
      }
      else{
        lcd.print(" -");   
      }
    }
    // CB
    if (CBswLst != CBsw) {
      CBswLst = CBsw;
      LCDchg = true;
      lcd.setCursor(0, 2);
     if (CBsw) {
        lcd.print("CB");
      }
      else {
        lcd.print("RX");
      }
    }

    // RX
    if (RXinxLst != RXinx) {
      RXinxLst = RXinx;
      LCDchg = true;
      lcd.setCursor(6, 2);
      if (RXinx > 15) {
        lcd.print(" unused");   //RX band
      }
      else {
        lcd.print( rxLCD[RXinx ] );   //RX band
      }
      lcd.setCursor(3, 2);
      if(RXinx<16){
        if((RXinx) < 10) lcd.print(" ");
        lcd.print(RXinx);   //RX band
      }
      else{
        lcd.print(" -");   
      }
    }

    if (RVvalLst != RVval) {
      RVvalLst = RVval;
      LCDchg = true;
      lcd.setCursor(0, 3);
      if (RVval != 99) {
        lcd.print("RV");
        lcd.setCursor(3,3);
        if (RVval < 10) lcd.print(" ");
        lcd.print(RVval);
        lcd.setCursor(6, 3);
        lcd.print(rvLCD[RVval]);
      }
      else {
        lcd.print("             ");
      }
    }

    // PTT
    if ((TXinpLst != TXinp)||(TXchgLst!=TXchg)) {
      TXinpLst  = TXinp;
      TXchgLst = TXchg;
      LCDchg = true;
      lcd.setCursor(15, 1);
      if (TXinp) {
        lcd.print("ON  ");
      }
      else {
        if(TXchg){
          lcd.print("hold");
        }
        else{
          lcd.print("off ");
        }
      }
    }
    if(overMinLst!=overMin){
      overMinLst=overMin;
      lcd.setCursor(14,3);
      if(overMin<100) lcd.print(" ");
      if(overMin< 10) lcd.print(" ");
      lcd.print(overMin);
    }
    if(overSecLst!=overSec){
      overSecLst=overSec;
      lcd.setCursor(18,3);
      if(overSec< 10) lcd.print("0");
      lcd.print(overSec);
    }

    if (!LCDchg) {
      get3V3();
    }

  } // end of MCPlog if at top of Loop
  
  if (MCPlog || (PD3V3 < 3000) ) {  // display errors
    CycRst();    // timing cycle reset
    if (!errLCD){
      lcd.clear();
      LCDrst();
      errLCD = true;
    }
    
    IO_MCP_INIT();
    if (MCPlog) {   // error still there
      lcd.setCursor(0, 0);
      lcd.print("IO expander error");
      lcd.setCursor(0, 1);
      lcd.print(bitRead(MCPlog, 7));
      lcd.print(bitRead(MCPlog, 6));
      lcd.print(bitRead(MCPlog, 5));
      lcd.print(bitRead(MCPlog, 4));
      lcd.print(bitRead(MCPlog, 3));
      lcd.print(bitRead(MCPlog, 2));
      lcd.print(bitRead(MCPlog, 1));
      lcd.print(bitRead(MCPlog, 0));
    } //local MCPlog

    get3V3();
    lcd.setCursor(0, 3);
    lcd.print("PD V");
    lcd.setCursor(5, 3);
    lcd.print(PD3V3);

#if RCVR
    lcd.setCursor(10, 3);
    lcd.print("RV V");
    lcd.setCursor(15, 3);
    lcd.print(RV3V3);
#endif

  }
  else{
    if (errLCD) {     // rewrite LCD text
      LCDrst();
      LCDbkgd();
      errLCD=false;
    }
  }

} // end loop

/*     **** time tick ****
        tick25 is turned ON by an interrupt every 25 mS.
        It's transferred to tick25A at the beginning of the loop
        and turned OFF after one scan. It can be used by a
        number of time delays.
*/

void timerIsr() {   // ****  interrupt service routine  ****
  tick25 = true;
}

void IO_MCP( byte MCPaddr, byte MCPreg, byte MCPdata) {
  Wire.beginTransmission(MCPaddr);
  Wire.write(MCPreg);
  Wire.write(MCPdata);
  MCPerr = Wire.endTransmission();
}

void IO_MCP_INIT() {
  IO_MCP(IOexp0, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp0, IODIRB, IOout);  //set port for out

#if Exp_21H
  IO_MCP(IOexp1, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp1, IODIRB, IOout);  //set port for out
#endif

#if Exp_22H
  IO_MCP(IOexp2, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp2, IODIRB, IOout);  //set port for out
#endif

#if Exp_23H
  IO_MCP(IOexp3, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp3, IODIRB, IOout);  //set port for out
#endif

#if Exp_24H
  IO_MCP(IOexp4, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp4, IODIRB, IOout);  //set port for out
#endif

#if Exp_25H
  IO_MCP(IOexp5, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp5, IODIRB, IOout);  //set port for out
#endif

#if Exp_26H
  IO_MCP(IOexp6, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp6, IODIRB, IOout);  //set port for out
#endif

#if Exp_27H
  IO_MCP(IOexp7, IODIRA, IOout);  //set port for out
  IO_MCP(IOexp7, IODIRB, IOout);  //set port for out
#endif

  for(byte j=0;j<16;j++){  //check invert bit, set IO data
    if(IOdef[j] [2]) IO_P[j]=0xFF; else IO_P[j]=0;
  }
  
  IO_MCP_WR();    //INIT all bits, check errors
}

void IO_MCP_WR() {
  // write to IO expanders, log errors
  MCPlog = 0;
  IO_MCP(IOexp0,  GPIOA, IO_P[0]);
  if (MCPerr) bitSet(MCPlog, 0);
  IO_MCP(IOexp0,  GPIOB, IO_P[1]);
  if (MCPerr) bitSet(MCPlog, 0);


#if Exp_21H
  IO_MCP(IOexp1,  GPIOA, IO_P[2]);
  if (MCPerr) bitSet(MCPlog, 1);
  IO_MCP(IOexp1,  GPIOB, IO_P[3]);
  if (MCPerr) bitSet(MCPlog, 1);
#endif

#if Exp_22H
  IO_MCP(IOexp2,  GPIOA, IO_P[4]);
  if (MCPerr) bitSet(MCPlog, 2);
  IO_MCP(IOexp2,  GPIOB, IO_P[5]);
  if (MCPerr) bitSet(MCPlog, 2);
#endif

#if Exp_23H
  IO_MCP(IOexp3,  GPIOA, IO_P[6]);
  if (MCPerr) bitSet(MCPlog, 3);
  IO_MCP(IOexp3,  GPIOB, IO_P[7]);
  if (MCPerr) bitSet(MCPlog, 3);
#endif

#if Exp_24H
  IO_MCP(IOexp4,  GPIOA, IO_P[8]);
  if (MCPerr) bitSet(MCPlog, 4);
  IO_MCP(IOexp4,  GPIOB, IO_P[9]);
  if (MCPerr) bitSet(MCPlog, 4);
#endif

#if Exp_25H
  IO_MCP(IOexp5,  GPIOA, IO_P[10]);
  if (MCPerr) bitSet(MCPlog, 5);
  IO_MCP(IOexp5,  GPIOB, IO_P[11]);
  if (MCPerr) bitSet(MCPlog, 5);
#endif

#if Exp_26H
  IO_MCP(IOexp6,  GPIOA, IO_P[12]);
  if (MCPerr) bitSet(MCPlog, 6);
  IO_MCP(IOexp6,  GPIOB, IO_P[13]);
  if (MCPerr) bitSet(MCPlog, 6);
#endif

#if Exp_27H
  IO_MCP(IOexp7,  GPIOA, IO_P[14]);
  if (MCPerr) bitSet(MCPlog, 7);
  IO_MCP(IOexp7,  GPIOB, IO_P[15]);
  if (MCPerr) bitSet(MCPlog, 7);
#endif
}

void LCDbkgd(){
  lcd.setCursor(15, 0);
  lcd.print("PTT");
  lcd.setCursor(17, 3);
  lcd.print(":");
  lcd.setCursor(0, 0);
  lcd.print("PD");
  lcd.setCursor(0, 1);
  lcd.print("TX");
  lcd.setCursor(0, 2);
  lcd.print("RX");
}

void LCDrst(){
  IO_MCP_INIT();
  lcd.clear();
  CBswLst   =  0;     // last - cross band enabled
  CBvalLst  = 99;     // last - cross band hex switch
  PDvalLst  = 99;     // last - Portsdown band bits
  RVvalLst  = 99;     // last - receiver band bits
  RXinxLst  = 99;     // last select RX or crossband LCD
  TXinxLst  = 99;
  TVinxLst  = 99;     // last selected transverter
  CBinxLst  =  0;     // last Cross band on
  TXinpLst  =  1;     // last TXinp
  overSecLst = 59;
  overMinLst = 99;
  }

 void get3V3() {
  L3V3 = (analogRead(T3V3));
  PD3V3 = (L3V3 * anaCal) >> 8; // volts times 1000
#if RCVR
  L3V3 = (analogRead(R3V3));
  RV3V3 = (L3V3 * anaCal) >> 8; // volts times 1000
#else
  RV3V3 = 0;
#endif
}

void CycRst() {
  T2Rtmr = T2Rdly+1;    // preset T2R delay timer
  R2Ttmr = R2Tdly+1;    // preset R2T delay timer
  T2R = false;
  R2T = false;
  TRrly = false;
  PPtmr = PPdly;
  PPOhold = false;
}

byte BitPat( byte IOport) {
  byte path;
  byte RXTX;
  byte BitW = 0;
  RXTX = IOdef[IOport] [0];  // get RX-TX type
  if(RXTX>6) return 0;

  if (TRout) {                    // TX on
    switch (RXTX) {
      case 0:       // RX  OFF
        return 0;
        break;
      case 1:       // RX  RX
        path = RXinx;
        break;
      case 2:       // OFF  TX
        path = TXinx;
        break;
      case 3:       // TX  TX
        path = TXinx;
        break;
      case 4:       // OFF PA-DC
        path = TXinx;
        break;
      case 5:       // OFF PA-PTT
        path = TXinx;
        break;
      case 6:       // RX  TX
        path = TXinx;
        break;
        default:
        return 0;
        break;
    }
  }
  else              // receiving
  {
    switch (RXTX) { // not TRout
      case 0:       // RX  OFF
        path = RXinx;
        break;
      case 1:       // RX  RX
        path = RXinx;
        break;
      case 2:       // OFF  TX
        return 0;
        break;
      case 3:       // TX  TX
        path = TXinx;
        break;
      case 4:       // OFF PA-DC
        path = TXinx;
        break;
      case 5:       // OFF PA-PTT
        path = TXinx;
        break;
      case 6:       // RX  TX
        path = RXinx;
        break;
        default:
        return 0;
        break;
    }
  }
    switch (RXTX) {   //timing
      case 0:         // RX  OFF
        //on during RX
        break;
      case 1:       // RX  RX
        //on during RX
        break;
      case 2:       // OFF  TX
        //return 0;
        break;
      case 3:       // TX  TX
        //on during TX
        break;
      case 4:       // OFF PA-DC
        if(!PApwr) return 0;
        break;
      case 5:       // OFF PA-PTT
        if(!PTT) return 0;
        break;
      case 6:       // RX  TX
        //on during RX and TX
        break;
        default:
        return 0;
        break;
    }
  if(path>15) return 0;
  switch(IOdef[IOport] [1]){    // bit mode
    case 0:                         // 0-7
      if(path<8) bitSet( BitW,path);
    break;
    case 1:                         // 8-15
      if(path>7) bitSet( BitW,path-8);
    break;
    case 2:                         // HEX + 0
      BitW = path;
    break;
    case 3:                         // HEX + 1
      BitW = path + 1;
    break;
    case 4:                         // 0-7 User 1
      BitW=lowByte(UDBP1[path]);
    break;
    case 5:                         // 8-15 User 1
      BitW=highByte(UDBP1[path]);
    break;
    case 6:                         // 0-7 User 2
      BitW=lowByte(UDBP2[path]);
    break;
    case 7:                         // 8-15 User 2
      BitW=highByte(UDBP2[path]);
    break;
    default:
    return 0;
    break;
  }
  return BitW;  
}
